# -*- coding: utf8 -*-

__title__ = "Amelet HDF5 exporter"
__author__ = "Guillaume PRIN"
__url__ = [""]


import os
import numpy
import tables
import FreeCAD
import re
if (int(FreeCAD.Version()[2].split()[0]) < 3702):
    from PyQt4 import QtGui, QtCore
else:
    from PySide import QtGui, QtCore
import Part
from Part import Edge
from Part import Line
from Part import Wire

FILTER = tables.Filters(complevel=5, complib='zlib')
RATIO = 1000.


#Called when freecad exports a file
def export(exportList, filename):
    doc = FreeCAD.ActiveDocument
    root_elements = []
    for child in FreeCAD.ActiveDocument.findObjects():
        if not(child.ViewObject.isVisible()):
            continue
        if len(child.InList) == 0:
            root_elements.append(child)

    #Remove hdf file if exists
    try:
        os.remove(filename)
    except OSError:
        pass

    #Create file
    h5file = tables.openFile(filename, mode="w", \
                                PYTABLES_SYS_ATTRS=False, filters=FILTER)
    h5file.setNodeAttr("/", 'FORMAT', 'AMELETHDF')
    h5file.setNodeAttr("/", 'AMELETHDF_FORMAT_VERSION', '1.5.4')
    meshgrp = h5file.createGroup("/", 'mesh', '')

    label = norm(doc.Label)
    FreeCAD.Console.PrintMessage("\nMesh :" + label)
    mesh_group = h5file.createGroup(meshgrp, label, '')
    h5file.setNodeAttr("/", 'entryPoint', "/mesh/" + label)
    hdf5_gmesh = GMesh(h5file, mesh_group, label)
    index_points = 0
    for element in root_elements:
        FreeCAD.Console.PrintMessage("\nExport Element " + str(element.TypeId))
        if not(element.ViewObject.isVisible()):
            FreeCAD.Console.PrintMessage("\n* #Not visible :" + norm(element.Label))
        elif is_group(element):
            index_points = export_list_groups(h5file, element, \
                                        hdf5_gmesh, index_points, 0)
        elif is_mesh(element):
            index_points = export_meshfaces(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, 0)
        elif is_helix(element):
            index_points = export_helix(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, 0)
        elif is_circle(element):
            index_points = export_circle(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, 0)
        elif is_ellipse(element):
            index_points = export_ellipse(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, 0)
        elif is_line(element):
            index_points = export_lines(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, 0)
        elif is_point(element):
            index_points = export_points(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, 0)
        elif is_edge(element):
            index_points = export_edges(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, 0)
        elif is_wire(element):
            index_points = export_wires(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, 0)
        elif is_group_points(element):
            index_points = export_group_points(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, 0)
        else:
            FreeCAD.Console.PrintMessage("\n* Unknow type for :" + norm(element.Label)\
                                            + " : " +  str(element.TypeId))
    h5file.close()
    FreeCAD.Console.PrintMessage("\nExport the HDF5 file is finished : '" + \
                                    filename + "'\n")


#Create gmesh and all the arrays
class GMesh():
    def __init__(self, h5file, mesh_group, label):
        self.gmesh_group = h5file.createGroup(mesh_group, \
                                        label, '')
        h5file.setNodeAttr(self.gmesh_group, 'type', 'unstructured')

        # Create nodes
        self.nodes_array = h5file.createEArray(self.gmesh_group, \
                                            "nodes", \
                                            tables.Float32Atom(), \
                                            (0, 3), filters=FILTER)
        self.elttype_array = h5file.createEArray(self.gmesh_group, \
                                            "elementTypes", \
                                            tables.UInt8Atom(), (0,), \
                                            filters=FILTER)
        self.eltnodes_array = h5file.createEArray(self.gmesh_group, \
                                            "elementNodes", \
                                            tables.UInt32Atom(), (0,), \
                                            filters=FILTER)
        h5file.createGroup(self.gmesh_group, 'selectorOnMesh')
        self.group_group = h5file.createGroup(self.gmesh_group, \
                                            "group")
        self.list_groups_group = h5file.createGroup(self.gmesh_group, \
                                            "groupGroup")


# Export a list of groups
def export_list_groups(h5file, groupGroup, hdf5_gmesh, \
                    index_points, indent):
    indent += 1
    FreeCAD.Console.PrintMessage("\n" + tabs(indent) + "Group :" + norm(groupGroup.Label))
    children = []
    for element in groupGroup.Group:
        if not(element.ViewObject.isVisible()):
            FreeCAD.Console.PrintMessage("\n" + tabs(indent) + "    Not visible :" + norm(element.Label))
            continue
        if is_group(element):
            index_points = export_list_groups(h5file, element, \
                                 hdf5_gmesh, index_points, \
                                 indent)
        elif is_mesh(element):
            index_points = export_meshfaces(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, indent)
        elif is_helix(element):
            index_points = export_helix(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, indent)
        elif is_circle(element):
            index_points = export_circle(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, indent)
        elif is_ellipse(element):
            index_points = export_ellipse(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, indent)
        elif is_line(element):
            index_points = export_lines(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, indent)
        elif is_point(element):
            index_points = export_points(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, indent)
        elif is_edge(element):
            index_points = export_edges(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, indent)
        elif is_wire(element):
            index_points = export_wires(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, 0)
        elif is_group_points(element):
            index_points = export_group_points(h5file, element, \
                                    hdf5_gmesh, \
                                    index_points, indent)
        else:
            FreeCAD.Console.PrintMessage("\n" + tabs(indent) + \
                    "Unknow type for :" + norm(element.Label) + " : " +  str(element.TypeId))
            continue
        children.append(norm(element.Label))
    if len(children) != 0:
        h5file.createArray(hdf5_gmesh.list_groups_group, \
                                    norm(groupGroup.Label), \
                                    children)
    return index_points


# Export mesh of triangles
def export_meshfaces(h5file, mesh, \
                    hdf5_gmesh, \
                    index_points, indent):
    "Export a group"
    indent += 1
    FreeCAD.Console.PrintMessage("\n" + tabs(indent) + "Mesh :" + norm(mesh.Label))
    group_mesh = h5file.createEArray(hdf5_gmesh.group_group, \
                                        norm(mesh.Label), \
                                        tables.UInt32Atom(), (0,), \
                                        filters=FILTER)
    for point in mesh.Mesh.Points:
        nodes = []
        nodes.append(numpy.float32(point.Vector[0] / RATIO))
        nodes.append(numpy.float32(point.Vector[1] / RATIO))
        nodes.append(numpy.float32(point.Vector[2] / RATIO))
        nodes_reshape = numpy.reshape(nodes, (1, 3))
        hdf5_gmesh.nodes_array.append(nodes_reshape)
    for facet in mesh.Mesh.Facets:
        hdf5_gmesh.elttype_array.append([11])
        group_mesh.append([hdf5_gmesh.elttype_array.nrows - 1])
        hdf5_gmesh.eltnodes_array.append(\
                                [facet.PointIndices[0] + index_points])
        hdf5_gmesh.eltnodes_array.append(\
                                [facet.PointIndices[1] + index_points])
        hdf5_gmesh.eltnodes_array.append(\
                                [facet.PointIndices[2] + index_points])

    h5file.setNodeAttr(group_mesh, "type", "element")
    h5file.setNodeAttr(group_mesh, "entityType", "face")
    index_points += mesh.Mesh.CountPoints
    return index_points


# Export line
def export_lines(h5file, line, \
                    hdf5_gmesh, \
                    index_points, indent):
    indent += 1
    FreeCAD.Console.PrintMessage("\n" + tabs(indent) + "Line :" + norm(line.Label))
    group_mesh = h5file.createEArray(hdf5_gmesh.group_group, \
                                        norm(line.Label), \
                                        tables.UInt32Atom(), (0,), \
                                        filters=FILTER)
    for edge in line.Shape.Edges:
        nodes = []
        point_0 = edge.Vertexes[0].Point
        point_1 = edge.Vertexes[1].Point
        nodes.append(numpy.float32(point_0.x / RATIO))
        nodes.append(numpy.float32(point_0.y / RATIO))
        nodes.append(numpy.float32(point_0.z / RATIO))
        nodes.append(numpy.float32(point_1.x / RATIO))
        nodes.append(numpy.float32(point_1.y / RATIO))
        nodes.append(numpy.float32(point_1.z / RATIO))
        nodes_reshape = numpy.reshape(nodes, (2, 3))
        hdf5_gmesh.nodes_array.append(nodes_reshape)
        index_points += 2
        hdf5_gmesh.elttype_array.append([1])
        group_mesh.append([hdf5_gmesh.elttype_array.nrows - 1])
        hdf5_gmesh.eltnodes_array.append([hdf5_gmesh.nodes_array.nrows - 2])
        hdf5_gmesh.eltnodes_array.append([hdf5_gmesh.nodes_array.nrows - 1])

    h5file.setNodeAttr(group_mesh, "type", "element")
    h5file.setNodeAttr(group_mesh, "entityType", "edge")
    return index_points

# Export a wire, for helix, circle, ellipse
def export_wire(h5file, vectors, group_mesh, hdf5_gmesh, index_points):
    for i in range(len(vectors)-1):
        nodes = []
        nodes.append(numpy.float32(vectors[i].x / RATIO))
        nodes.append(numpy.float32(vectors[i].y / RATIO))
        nodes.append(numpy.float32(vectors[i].z / RATIO))
        nodes.append(numpy.float32(vectors[i + 1].x / RATIO))
        nodes.append(numpy.float32(vectors[i + 1].y / RATIO))
        nodes.append(numpy.float32(vectors[i + 1].z / RATIO))
        nodes_reshape = numpy.reshape(nodes, (2, 3))
        hdf5_gmesh.nodes_array.append(nodes_reshape)
        index_points += 2
        hdf5_gmesh.elttype_array.append([1])
        group_mesh.append([hdf5_gmesh.elttype_array.nrows - 1])
        hdf5_gmesh.eltnodes_array.append([hdf5_gmesh.nodes_array.nrows - 2])
        hdf5_gmesh.eltnodes_array.append([hdf5_gmesh.nodes_array.nrows - 1])
    
    h5file.setNodeAttr(group_mesh, "type", "element")
    h5file.setNodeAttr(group_mesh, "entityType", "edge")
    return index_points

# Export helix
def export_helix(h5file, helix, \
                    hdf5_gmesh, \
                    index_points, indent):
    indent += 1
    FreeCAD.Console.PrintMessage("\n" + tabs(indent) + "Helix :" + norm(helix.Label))
    wire=helix.Shape
    vectors = helix.Shape.discretize(int(float(helix.Height) / helix.Pitch * 14))
    group_mesh = h5file.createEArray(hdf5_gmesh.group_group, \
                                        norm(helix.Label), \
                                        tables.UInt32Atom(), (0,), \
                                        filters=FILTER)
    return export_wire(h5file, vectors, group_mesh, hdf5_gmesh, index_points)

# Export circle
def export_circle(h5file, circle, \
                    hdf5_gmesh, \
                    index_points, indent):
    indent += 1
    FreeCAD.Console.PrintMessage("\n" + tabs(indent) + "Circle :" + norm(circle.Label))
    wire=circle.Shape
    vectors = circle.Shape.discretize(int(abs(circle.Angle1 - circle.Angle0) * 100))
    group_mesh = h5file.createEArray(hdf5_gmesh.group_group, \
                                        norm(circle.Label), \
                                        tables.UInt32Atom(), (0,), \
                                        filters=FILTER)
    return export_wire(h5file, vectors, group_mesh, hdf5_gmesh, index_points)

# Export ellipse
def export_ellipse(h5file, ellipse, \
                    hdf5_gmesh, \
                    index_points, indent):
    indent += 1
    FreeCAD.Console.PrintMessage("\n" + tabs(indent) + "Ellipse :" + norm(ellipse.Label))
    wire=ellipse.Shape
    vectors = ellipse.Shape.discretize(int(abs(ellipse.Angle1 - ellipse.Angle0) / 10))
    group_mesh = h5file.createEArray(hdf5_gmesh.group_group, \
                                        norm(ellipse.Label), \
                                        tables.UInt32Atom(), (0,), \
                                        filters=FILTER)
    return export_wire(h5file, vectors, group_mesh, hdf5_gmesh, index_points)


# Export points
def export_points(h5file, point, \
                    hdf5_gmesh, \
                    index_points, indent):
    indent += 1
    FreeCAD.Console.PrintMessage("\n" + tabs(indent) + "Point :" + norm(point.Label))
    group_mesh = h5file.createEArray(hdf5_gmesh.group_group, \
                                        norm(point.Label), \
                                        tables.UInt32Atom(), (0,), \
                                        filters=FILTER)
    nodes = []
    nodes.append(numpy.float32(point.X / RATIO))
    nodes.append(numpy.float32(point.Y / RATIO))
    nodes.append(numpy.float32(point.Z / RATIO))
    nodes_reshape = numpy.reshape(nodes, (1, 3))
    hdf5_gmesh.nodes_array.append(nodes_reshape)
    index_points += 1

    group_mesh.append([hdf5_gmesh.nodes_array.nrows - 1])
    h5file.setNodeAttr(group_mesh, "type", "node")
    return index_points

# Export edge
def export_edges(h5file, edge, \
                    hdf5_gmesh, \
                    index_points, indent):
    indent += 1
    FreeCAD.Console.PrintMessage("\n" + tabs(indent) + "Edge :" + norm(edge.Label))
    if type(edge.Shape.Curve) is Line:
        group_mesh = h5file.createEArray(hdf5_gmesh.group_group, \
                                            norm(edge.Label), \
                                            tables.UInt32Atom(), (0,), \
                                            filters=FILTER)
        nodes = []
        point_0 = edge.Shape.Vertexes[0].Point
        point_1 = edge.Shape.Vertexes[1].Point
        nodes.append(numpy.float32(point_0.x / RATIO))
        nodes.append(numpy.float32(point_0.y / RATIO))
        nodes.append(numpy.float32(point_0.z / RATIO))
        nodes.append(numpy.float32(point_1.x / RATIO))
        nodes.append(numpy.float32(point_1.y / RATIO))
        nodes.append(numpy.float32(point_1.z / RATIO))
        nodes_reshape = numpy.reshape(nodes, (2, 3))
        hdf5_gmesh.nodes_array.append(nodes_reshape)
        index_points += 2
        hdf5_gmesh.elttype_array.append([1])
        group_mesh.append([hdf5_gmesh.elttype_array.nrows - 1])
        hdf5_gmesh.eltnodes_array.append([hdf5_gmesh.nodes_array.nrows - 2])
        hdf5_gmesh.eltnodes_array.append([hdf5_gmesh.nodes_array.nrows - 1])

        h5file.setNodeAttr(group_mesh, "type", "element")
        h5file.setNodeAttr(group_mesh, "entityType", "edge")
        return index_points
    else:
        p0 = edge.Shape.Vertexes[0].Point
        p1 = edge.Shape.Vertexes[1].Point
        nb = p0.distanceToPoint(p1) / edge.Shape.Curve.length() * 60
        vectors = edge.Shape.discretize(int(nb))
        group_mesh = h5file.createEArray(hdf5_gmesh.group_group, \
                                            norm(edge.Label), \
                                            tables.UInt32Atom(), (0,), \
                                            filters=FILTER)
        return export_wire(h5file, vectors, group_mesh, hdf5_gmesh, index_points)

# Export edge
def export_wires(h5file, wire, \
                    hdf5_gmesh, \
                    index_points, indent):
    indent += 1
    FreeCAD.Console.PrintMessage("\n" + tabs(indent) + "Wire :" + norm(wire.Label))
    p0 = wire.Shape.Vertexes[0].Point
    p1 = wire.Shape.Vertexes[1].Point
    nb = p0.distanceToPoint(p1) / wire.Shape.Length * 30
    vectors = wire.Shape.discretize(int(nb))
    group_mesh = h5file.createEArray(hdf5_gmesh.group_group, \
                                        norm(wire.Label), \
                                        tables.UInt32Atom(), (0,), \
                                        filters=FILTER)
    return export_wire(h5file, vectors, group_mesh, hdf5_gmesh, index_points)

# Export group points
def export_group_points(h5file, group_points, \
                    hdf5_gmesh, \
                    index_points, indent):
    indent += 1
    FreeCAD.Console.PrintMessage("\n" + tabs(indent) + "Group Points :" + norm(group_points.Label))
    group_mesh = h5file.createEArray(hdf5_gmesh.group_group, \
                                        norm(group_points.Label), \
                                        tables.UInt32Atom(), (0,), \
                                        filters=FILTER)
    nodes = []
    for point in group_points.Group:
        FreeCAD.Console.PrintMessage("\n" + tabs(indent + 1) + "Points :" + norm(point.Label))
        nodes = []
        nodes.append(numpy.float32(point.X / RATIO))
        nodes.append(numpy.float32(point.Y / RATIO))
        nodes.append(numpy.float32(point.Z / RATIO))
        nodes_reshape = numpy.reshape(nodes, (1, 3))
        hdf5_gmesh.nodes_array.append(nodes_reshape)
        index_points += 1
        group_mesh.append([hdf5_gmesh.nodes_array.nrows - 1])

    h5file.setNodeAttr(group_mesh, "type", "node")
    return index_points


#Convert a number in indentation
def tabs(indent):
    s = "-"
    for i in range(indent):
        s += "-"
    return s


#Is a group
def is_group(element):
    return element.TypeId == 'App::DocumentObjectGroup' and \
                        not(is_group_points(element))


#Is a group of points
def is_group_points(element):
    if element.TypeId != 'App::DocumentObjectGroup':
        return False
    regex = "(" + norm(element.Label) + ")(_)(\\d+)"
    rg = re.compile(regex, re.IGNORECASE | re.DOTALL)
    for child in element.Group:
        if child.TypeId != 'Part::Vertex':
            return False
        if not(rg.search(norm(child.Label))):
            return False
    return True


#Is a mesh
def is_mesh(element):
    return element.TypeId == 'Mesh::Feature'


#Is line
def is_line(element):
    return element.TypeId == 'Part::Line' or \
            element.TypeId == 'Part::Part2DObjectPython' or \
            element.TypeId == 'Sketcher::SketchObject' or \
            element.TypeId == 'Part::RegularPolygon'

#Is helix
def is_helix(element):
    return element.TypeId == 'Part::Helix'

#Is circle
def is_circle(element):
    return element.TypeId == 'Part::Circle'

#Is ellipse
def is_ellipse(element):
    return element.TypeId == 'Part::Ellipse'

#Is point
def is_point(element):
    return element.TypeId == 'Part::Vertex'

#Is edge
def is_edge(element):
    return element.TypeId == 'Part::Feature' and \
           isinstance(element.Shape, Edge)
#Is wire
def is_wire(element):
    return element.TypeId == 'Part::Feature' and \
           isinstance(element.Shape, Wire)

#Normalize label remove such thing as '(Meshed)'
def norm(label):
    if '(' in str(label):
        return str(label).split('(')[0].strip()
    return str(label).strip()


#Dialog message
def message_dialog(title, msg):
    diag = QtGui.QMessageBox(QtGui.QMessageBox.Information, title, msg)
    diag.setWindowModality(QtCore.Qt.ApplicationModal)
    diag.exec_()

#End
